# -*- coding: utf-8 -*-

import os
import json
import numpy as np
from glob import glob
from collections import defaultdict

import tensorflow as tf
from bert4keras.tokenizers import Tokenizer

from cwseg.config import config
from cwseg.download import download_model

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class TrieNode(object):
    """
    defaultdict构造trie前缀树
    """

    def __init__(self):
        # 当前节点的子树节点
        self.children = defaultdict(TrieNode)
        # 从根节点到该节点的路径是否一个单词
        self.is_word = False


class Trie(object):

    def __init__(self):
        """
        Initialize your data structure here.
        """
        # 字典树的根节点
        self.root = TrieNode()

    def insert(self, word: str):
        """
        Inserts a word into the trie.
        :param word: str
        :return: None
        """
        # 从根节点开始遍历
        cur_node = self.root
        # 每个字符是一个节点，如果没有则新建，有则直接向下遍历继续插入
        for ch in word:
            cur_node = cur_node.children[ch]
        # 整个单词中的每个字符都对应有一个节点了，并且是通过children关联起来的，
        # 那么这最后一个字符的节点就表示一个单词的结尾
        cur_node.is_word = True

    def search(self, word: str):
        """
        Returns if the word is in the trie.
        :param word: str
        :return: bool
        """
        # 从根节点向下遍历
        cur_node = self.root
        for ch in word:
            cur_node = cur_node.children.get(ch)
            # 如果某个字符没有对应的节点，说明其对应的单词没有在该字典树中
            if not cur_node:
                return False
        # 该单词所有的字符均有对应的节点，那么判断最后一个字符是否是字典中一个单词的结尾
        return cur_node.is_word

    def startsWith(self, prefix):
        """
        Returns if there is any word in the trie that starts with the given prefix.
        :param word: str
        :return: bool
        """
        # 从根节点向下遍历
        cur_node = self.root
        for ch in prefix:
            # 如果某个字符没有对应的节点，说明其对应的单词没有在该字典树中
            cur_node = cur_node.children.get(ch)
            if not cur_node:
                return False
        # 该单词所有的字符均有对应的节点，那么该单词一个是字典树中某些单词的前缀
        return True


class TrieSeg(object):
    def __init__(self, vocab_set):
        self.vocab_set = vocab_set
        self.trie = Trie()  # 构建前缀树词典，提高查询效率
        [self.trie.insert(word) for word in self.vocab_set]
        self.max_len = max([len(word) for word in self.vocab_set])

    def forword_max_match(self, sent):
        """
        前向最大匹配，依次移动游标查询当前词是否在前缀树中；
        :param sent:  sent
        :return seg: seg for sent
        """
        start = 0
        seg = []
        while start <= len(sent) - 1:
            index = min(start + self.max_len, len(sent))
            for i in range(self.max_len):
                if index <= start:
                    # word not found
                    self.trie.insert(sent[start:start + 1])
                    break
                if self.trie.search(sent[start:index]):
                    seg.append(sent[start:index])
                    start = index
                    break
                index += -1
        return seg

    def forword_max_match_merge(self, sent, search_mode=False):
        """
        前向最大匹配，依次移动游标查询当前词是否在前缀树中；
        :param sent:  sent
        :return seg: seg for sent
        """
        start = 0
        seg = []
        while start < len(sent):
            # index = len(sent)
            for index in range(len(sent), start - 1, -1):
                # if the word was not found in the trie, then add it
                if index == start:
                    self.trie.insert(sent[start:start + 1])
                    seg.append(''.join(sent[start:start + 1]))
                    start += 1
                    break
                # found the words seq, merge
                if self.trie.search(''.join(sent[start:index])):
                    if search_mode and index - start > 1:  # search mode
                        seg.extend(sent[start:index])
                    seg.append(''.join(sent[start:index]))
                    start = index
                    break
        return seg

    def backword_max_match(self, sent):
        """
        后向最大匹配，依次移动游标查询当前词是否在前缀树中；
        :param sent:  sent
        :return seg: seg for sent
        """
        start = len(sent)
        seg = []
        while start > 0:
            index = start - self.max_len
            if index < 0:
                index = 0
            for i in range(self.max_len):
                if self.trie.search(sent[index:start]) or len(sent[index:start]) == 1:
                    seg.append(sent[index:start])
                    start = index
                    break
                index += 1
        return seg


def softmax(x, axis=-1):
    """numpy-softmax
    """
    x = x - x.max(axis=axis, keepdims=True)
    x = np.exp(x)
    return x / x.sum(axis=axis, keepdims=True)


class SavedModel:
    def __init__(self, modelpath, model_name, pattern='variables'):
        ## load saved model
        self.pattern = pattern
        self.modelpath = modelpath
        self.model_name = model_name
        self.paths = glob(os.path.join(self.modelpath, '{}'.format(self.pattern)))  ## 'modelname/version/variables'
        self.models = {}
        self.bertParam = {'tokenizer': {}, 'trans': {}, 'categories': {}}

        for p in self.paths:
            self.handle_model(p, status='create')

    def predict(self, inputs=None, inputNames=None, outNames=None, modelName=None):

        if type(inputNames) is not list:
            inputNames = [inputNames]
            inputs = [inputs]
        version = 'max'
        model = self.models.get(modelName, {}).get(version)
        if model is not None:
            infer = model.signatures["serving_default"]
            inputsDict = {}
            for ind, key in enumerate(inputs):
                inputsDict[inputNames[ind]] = tf.constant(inputs[ind])
            pred = infer(**inputsDict)
            if type(outNames) is str:
                pred = pred.get(outNames).numpy()
            elif type(outNames) is list:
                pred = [pred.get(name).numpy() for name in outNames]
            else:
                pred = None
            return pred
        else:
            return None

    def handle_model(self, path, status='create'):

        if status == 'create':
            self.load_model(path)

        elif status == 'delete':
            self.delete_model(path)

        self.update_new_version()

    def load_bert(self, path, modelName):
        ## transition matrix
        transPath = os.path.join(path, 'trans.json')
        if os.path.exists(transPath):
            try:
                with open(transPath) as f:
                    self.bertParam['trans'][modelName] = json.load(f)
            except:
                print('load trans error')

        ## tokenizer
        vocabPath = os.path.join(path, 'vocab.txt')
        if os.path.exists(vocabPath):
            try:
                # tokenizer = Tokenizer(vocabPath, do_lower_case=True)
                tokenizer = Tokenizer(vocabPath, do_lower_case=True, pre_tokenize=lambda x: list(x))
                self.bertParam['tokenizer'][modelName] = tokenizer
            except:
                print('load tokenizer error')

        ## category
        categoryPath = os.path.join(path, 'category.json')
        if os.path.exists(categoryPath):
            try:
                with open(categoryPath) as f:
                    self.bertParam['categories'][modelName] = json.load(f)
            except:
                print('load categories error')

    def load_model(self, path):
        ##
        if os.path.isdir(path) and path.startswith(self.modelpath) and len(path) > len(self.modelpath):
            try:
                modelWeights = path.rstrip(self.pattern)
                modelName, modelVersion = self.model_name, '0'
                if modelName not in self.models:
                    self.models[modelName] = {}
                if modelVersion not in self.models[modelName]:
                    model = tf.saved_model.load(modelWeights)
                    self.models[modelName][modelVersion] = model
                    self.load_bert(modelWeights, modelName)
                    print(f'load model: {modelName}, version: {modelVersion}')
            except:
                pass

    def update_new_version(self):
        for modelName in self.models:
            currentVersion = '0'
            model = self.models[modelName].get(currentVersion)
            for version in self.models[modelName]:
                if version.startswith('max'):
                    continue
                try:
                    if int(version) > int(currentVersion):
                        model = self.models[modelName][version]
                        currentVersion = version
                except:
                    pass
            if model is not None:
                self.models[modelName]['max'] = model

    def delete_model(self, path):
        if os.path.isdir(path):
            modelPaths = glob(os.path.join(path, '*', self.pattern))
            if len(modelPaths) > 0:
                ##
                modelName = path.split(os.path.sep)[-1]
                if modelName in self.models:
                    self.models.pop(modelName)
                    # logger.info(f'delete model {modelName}')

            else:
                modelPaths = os.path.join(path, self.pattern)
                if os.path.exists(modelPaths):
                    modelName, modelVersion = modelPaths.split(os.path.sep)[-3:-1]
                    if modelName in self.models:
                        if modelVersion in self.models[modelName]:
                            self.models[modelName].pop(modelVersion)
                            if len(self.models[modelName]):
                                self.models.pop(modelName)
                            # logger.info(f'delete model {modelName}, version: {modelVersion}')


class cwseg(object):
    """
    sequence labeler inference
    """

    def __init__(self,
                 model_name="bert-base-2",
                 user_dict="default",
                 starts=[0],
                 ends=[0]):

        """初始化函数，加载模型及用户词典"""
        if model_name in ["bert-base-2"]:
            config.modelDir = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "released_models",
                model_name,
            )
        elif model_name in config.available_models:
            config.modelDir = os.path.join(
                config.cwseg_home,
                model_name,
            )
            download_model(config.model_urls[model_name], config.cwseg_home, config.model_hash[model_name])
        else:
            print('Please check the model name!')

        model = SavedModel(config.modelDir, model_name)
        read_server = model.predict

        self.trans = model.bertParam['trans'][model_name]
        self.num_labels = len(self.trans)
        self.non_starts = []
        self.non_ends = []
        self.tokenizer = model.bertParam['tokenizer'][model_name]
        self.categories = model.bertParam['categories'][model_name]
        self.bertName = model_name
        self.model = read_server
        if starts is not None:
            for i in range(self.num_labels):
                if i not in starts:
                    self.non_starts.append(i)
        if ends is not None:
            for i in range(self.num_labels):
                if i not in ends:
                    self.non_ends.append(i)

        # build trie with dict
        vocab_set = self.load_vocab_set(user_dict)
        self.trie_seg = TrieSeg(set(vocab_set))

    def load_vocab_set(self, user_dict):
        if user_dict in ["default"]:
            vocab_file = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "dicts",
                "vocab_set.json",
            )
        elif user_dict in config.models_with_dict:
            vocab_file = os.path.join(
                config.cwseg_home,
                user_dict,
            )
            download_model(config.dict_urls[user_dict], config.cwseg_home, config.dict_hash[user_dict])
        elif user_dict[-4:] == 'json':
            vocab_file = user_dict
        else:
            print('Please check the dict name, or set your own dict path!')

        with open(vocab_file) as fin:
            for line in fin:
                new_vocab_set = json.loads(line)
        return new_vocab_set

    def decode(self, nodes):
        """nodes.shape=[seq_len, num_labels]
        """

        nodes[0, self.non_starts] -= np.inf
        nodes[-1, self.non_ends] -= np.inf

        # dp(viterbi)
        labels = np.arange(self.num_labels).reshape((1, -1))
        scores = nodes[0].reshape((-1, 1))
        paths = labels
        for l in range(1, len(nodes)):
            M = scores + self.trans + nodes[l].reshape((1, -1))
            idxs = M.argmax(0)
            scores = M.max(0).reshape((-1, 1))
            paths = np.concatenate([paths[:, idxs], labels], 0)

        # optimal path

        return paths[:, scores[:, 0].argmax()]

    def inference(self, text):
        if len(text) == 0:
            return []
        convertText = text[:]

        tokens = ['[CLS]'] + list(text.lower()[:510]) + ['[SEP]']
        tokens = ['[UNK]' if tk not in self.tokenizer._token_dict else tk for tk in tokens]
        mapping = self.tokenizer.rematch(convertText, tokens)
        nmap = len(mapping)
        for i in range(nmap):
            if len(mapping[i]) == 0 and i not in [0, nmap - 1]:
                ## white space
                mapping[i] = [i - 1]

        token_ids = self.tokenizer.tokens_to_ids(tokens)
        segment_ids = [0] * len(token_ids)
        token_ids, segment_ids = np.array([token_ids], dtype='float32'), np.array([segment_ids], dtype='float32')
        pred = self.model([token_ids, segment_ids], ['Input-Token', 'Input-Segment'], 'dense', self.bertName)
        nodes = pred[0]

        labels = self.decode(nodes)
        probs = softmax(nodes)
        entities, starting = [], False
        for i, label in enumerate(labels):
            if label > 0:
                if label % 2 == 1:
                    starting = True
                    entities.append([[i], self.categories[(label - 1) // 2]])
                elif starting:
                    entities[-1][0].append(i)
                else:
                    starting = False
            else:
                starting = False
        res = [(mapping[w[0]][0], mapping[w[-1]][-1], l, probs[w[-1], 1]) for w, l in entities]
        res = [[itm[0], itm[1], text[itm[0]:itm[1] + 1], round(float(itm[-1]), 2), itm[2]] for itm in res]
        res = sorted(res, key=lambda x: x[0])  ## [begin, end, token, prob, category]
        ## map to text
        ntext = len(convertText)
        newres = []
        for line in res:
            if len(newres) == 0:
                newres.append(line)
            else:
                if line[0] - newres[-1][1] > 1:
                    begin = newres[-1][1] + 1
                    end = line[0]
                    newres.append([begin, end, convertText[begin:end], 1, 'cws'])
                newres.append(line)
        if len(newres) == 0:
            return newres
        if newres[-1][1] < ntext - 1:
            begin = newres[-1][1] + 1
            end = ntext - 1
            newres.append([begin, end, convertText[begin:end], 1, 'cws'])

        return newres

    def cut(self, text, cut_mode='fine'):
        '''

        Args:
            text:
            cut_mode: default is fine mode, ['fine', 'coarse', 'search']

        Returns:

        '''
        seg = [i[2] for i in self.inference(text)]

        if cut_mode == 'fine':
            return seg
        elif cut_mode == 'coarse':
            seg = self.trie_seg.forword_max_match_merge(seg, search_mode=False)
            return seg
        elif cut_mode == 'search':
            seg = self.trie_seg.forword_max_match_merge(seg, search_mode=True)
            return seg

    def cut_f(self, file, cut_mode='fine'):
        '''

        Args:
            file: file path, txt format
            cut_mode: default is fine mode, ['fine', 'coarse']

        Returns:
            *pred.txt
        '''
        pred_file = file.strip('txt') + 'pred.txt'
        with open(file) as fin, open(pred_file, 'w') as fout:
            for line in fin:
                text = line.strip()
                new_line = ''
                while len(text) > 0:
                    text_seg = text[:500]
                    seg = self.cut(text_seg, cut_mode=cut_mode)
                    new_line += ' '.join(seg)
                    text = text[500:]
                fout.write(new_line + '\n')
        print('Save cut results to file %s' % pred_file)


if __name__ == "__main__":
    cwseg = cwseg()
    print(cwseg.cut("北京清华大学"))
